#' @title main function for Reliability_MA
#' @description Please input data, and it'll output a reliability estimate for each observation and an estimate of the system reliability function.
#' @param data T observations, with rows denoting observations and columns denoting nodes.
#' @return reliability_observation Reliability estimate for each observation.
#' @return reliability_function Estimate of the system reliability function, where time t is an isometric series with the minimum value in data as the first term and the maximum value in data as the last term, totalling 50.
#' @import copula
#' @import VineCopula
#' @import Rsolnp
#' @export 


data <- read.csv("./water.csv")
Reliability_MA = function(data){
  library(copula)
  library(VineCopula)
  library(Rsolnp)
  set.seed(5)
  X = as.matrix(data)
  n = ncol(X)
  k = 6
  J = 10
  t0 = apply(X,1,sort)[(n-k+1),]
  samplesize = nrow(data)
  UX = pobs(X)
  EmCDF = function(x){
    c0 = matrix(nrow = nrow(x), ncol = 1)
    for(ii in 1:nrow(x)){
      eee = c()
      for(jj in 1:nrow(x)){
        ee <- x[jj,]<=x[ii,]
        eee = rbind(eee,ee)
      }
      eee <- eee+0
      eeee = apply(eee,1,prod)
      c0[ii] = sum(eeee)/(nrow(x))
    }
    return(c0)
  }
  fx = function(x,t){
    h = 1.06*min(sd(x),as.numeric(quantile(x,c(1/4,3/4))[2]-quantile(x,c(1/4,3/4))[1])/1.34)*((samplesize)^(-0.2))
    c0 = matrix(nrow = length(t),ncol=1)
    for(ii in 1:length(t)){
      c0[ii,1] = sum(abs(x-t[ii])<=h)/(2*samplesize*h)
    }
    return(c0)
  }
  
  Ut = matrix(nrow = length(t0),ncol = n)
  for(i in 1:n){
    for(ii in 1:length(t0)){
      e1 <- t0[ii]>=X[,i]
      e1 = e1+0
      Ut[ii,i] = sum(e1)/(samplesize)
    }
  }
  
  R1 = c()
  R0 = apply(X,1,sort)[(n-k+1),]
  for(i in 1:length(t0)){
    e2 <- t0[i]>=R0
    e2 = e2+0
    R1[i] = 1-sum(e2)/(samplesize)
  }
  
  surv_Copula = function(Ut0,copula){
    sc = 1
    for(j in 1:ncol(Ut0)){
      aco = combn(1:ncol(Ut0),j)
      sc_j = 0
      for(l in 1:(ncol(aco))){
        Ut1 = Ut0
        Ut1[,-c(aco[,l])] = 0.99999
        sc_j = sc_j+pCopula(Ut1,copula = copula)
      }
      sc_j = sc_j*((-1)^(j))
      sc = sc+sc_j
    }
    return(sc)
  }
  Rt = function(Ut,copula){
    rt=0
    for(j in k:n){
      aco=combn(1:n,j)
      rt_j=0
      for(l in 1:(ncol(aco))){
        Ut0=Ut
        Ut0[,-c(aco[,l])]=0
        rt_j=rt_j+surv_Copula(Ut0,copula)
      }
      rt_j=rt_j*((-1)^(j-k))*(choose(j-1,k-1))
      rt=rt+rt_j
    }
    return(rt)
  }
  
  candidate_model=list(gumbelCopula,joeCopula,frankCopula,normalCopula)
  M=length(candidate_model)
  candid_param=matrix(nrow = length(candidate_model),ncol = 1)
  param=matrix(nrow = length(candidate_model),ncol = 1)
  param_can=c(3,3,3,0.2)
  candid_MLE=function(X){
    for(m in 1:length(candidate_model)){
      model_fit=fitCopula(candidate_model[[m]](dim=n),data=X)
      param[m,1]=model_fit@estimate
    }
    return(param)
  }
  options (warn = -1)
  CV_out=matrix(nrow = samplesize,ncol = (M+1))
  w=runif(M,0,1)
  w=w/sum(w)
  for (j in 1:J) {
    id.test <- (floor((j-1)*samplesize/J)+1) : ceiling(j*samplesize/J)
    id.train <- setdiff(1:samplesize, id.test)
    X.train <- UX[id.train, ]
    X.test <- UX[id.test, ]
    param_try=try(candid_MLE(X.train))
    if("try-error" %in% class(param_try)){
      result=rep(NA,70)
      return(result)
      stop()
    }else{
      param=param_try
    }
    
    hatf_test=matrix(nrow = length(id.test),ncol=n)
    for(i in 1:n){
      hatf_test[,i]=fx(X[,i],X[id.test,i])
    }
    mix_list=list()
    for(m in 1:length(candidate_model)){
      CV_out[id.test,m]=dCopula(X.test,candidate_model[[m]](param = param[m,1],dim=n))
      mix_list[[m]]=candidate_model[[m]](dim=n)
    }
    
    mix_mode=mixCopula(mix_list,w=w)
    mix_solnp=function(param_mix){
      -loglikCopula(param =param_mix ,u=X.train,copula = mix_mode)
    }
    alphaw_mix=c()
    LB=c()
    UB=c()
    for(m in 1:M){
      a=candidate_model[[m]](dim=n)
      if(a@param.upbnd==Inf){
        alphaw_mix=c(alphaw_mix,runif(1,a@param.lowbnd,10))#2
        LB=c(LB,a@param.lowbnd)
        UB=c(UB,100)
      }else{
        alphaw_mix=c(alphaw_mix,runif(1,a@param.lowbnd,a@param.upbnd-0.001))#-0.45
        LB=c(LB,a@param.lowbnd)
        UB=c(UB,a@param.upbnd-0.01)
      }
    }
    LB=c(LB,rep(0,M))
    UB=c(UB,rep(1,M))
    alphaw_mix=c(alphaw_mix,runif(M,0,1))
    alphaw_mix[(M+1):(2*M)]=alphaw_mix[(M+1):(2*M)]/sum(alphaw_mix[(M+1):(2*M)])
    eq_mix=function(alphaw_mix) sum(alphaw_mix[(M+1):(2*M)])
    ret_mix=constrOptim(theta = alphaw_mix,f=mix_solnp,grad=NULL,
                        ui=rbind(c(1,rep(0,7)),c(0,1,rep(0,6)),c(0,0,1,rep(0,5)),c(0,0,0,1,rep(0,4)),
                                 c(0,0,0,-1,rep(0,4)),c(rep(0,4),1,0,0,0),c(rep(0,5),1,0,0),c(rep(0,6),1,0),
                                 c(rep(0,7),1),c(rep(0,4),rep(1,4)),c(rep(0,4),rep(-1,4))),
                        ci=c(1,1,0,-0.5,-0.9999,0,0,0,0,0.999,-1.001),outer.iterations = 10)
    hatalphaw_mix=ret_mix$par
    
    mix_list_hat=list()
    for(m in 1:length(candidate_model)){
      mix_list_hat[[m]]=candidate_model[[m]](hatalphaw_mix[m], dim=n)
    }
    
    mix_model=mixCopula(mix_list_hat, w=hatalphaw_mix[(M+1):(2*M)])
    CV_out[id.test,(M+1)]=dCopula(X.test,mix_model)
  }
  ob_fun=function(w){
    -sum(log(CV_out%*%w))
  }
  w1=runif(M+1,0,1)
  w1=w1/sum(w1)
  eq=function(w1) sum(w1)
  ret=solnp(w1,ob_fun,eqfun = eq,eqB = 1,LB=rep(0,M+1),UB=rep(1,M+1))
  hatw=ret$pars
  
  candid_param=candid_MLE(UX)
  mix_solnp=function(param_mix){
    -loglikCopula(param =param_mix ,u=UX,copula = mix_mode)
  }
  eq_mix=function(alphaw_mix) sum(alphaw_mix[(M+1):(2*M)])
  
  ret_mix=constrOptim(theta = alphaw_mix,f=mix_solnp,grad=NULL,
                      ui=rbind(c(1,rep(0,7)),c(0,1,rep(0,6)),c(0,0,1,rep(0,5)),c(0,0,0,1,rep(0,4)),
                               c(0,0,0,-1,rep(0,4)),c(rep(0,4),1,0,0,0),c(rep(0,5),1,0,0),c(rep(0,6),1,0),
                               c(rep(0,7),1),c(rep(0,4),rep(1,4)),c(rep(0,4),rep(-1,4))),
                      ci=c(1,1,0,-0.5,-0.9999,0,0,0,0,0.999,-1.001),outer.iterations = 10)
  hatalphaw_mix=ret_mix$par
  
  mix_list_3=list()
  for(m in 1:length(candidate_model)){
    mix_list_3[[m]]=candidate_model[[m]](hatalphaw_mix[m], dim=n)
  }
  mix_model=mixCopula(mix_list_3, w=hatalphaw_mix[(M+1):(2*M)])
  
  hat_para_list=list()
  for(m in 1:length(candidate_model)){
    hat_para_list[[m]]=candidate_model[[m]](candid_param[m], dim=n)
  }
  hat_MA_list=hat_para_list
  hat_MA_list[[M+1]]=mix_model
  MA_resu =Rt(Ut=Ut,copula = mixCopula(hat_MA_list,w=hatw))
  
  
  #t0=seq(0,0.8,length.out=50)
  t0=seq(min(X),max(X),length.out=50)
  Ut=matrix(nrow = length(t0),ncol = n)
  for(i in 1:n){
    for(ii in 1:length(t0)){
      e1<- t0[ii]>=X[,i]
      e1=e1+0
      Ut[ii,i]=sum(e1)/(samplesize)
    }
  }
  MA_resu_t = Rt(Ut=Ut,copula = mixCopula(hat_MA_list,w=hatw))
  return(list(reliability_observation=MA_resu, reliability_function=MA_resu_t))
}

out = Reliability_MA(data = data)
